/***********************************************************************
 *  i386-supp.c
 *
 *  Description:  Support functions for the i386 GDB target stub.
 *
 *  Credits:      Created by Jonathan Brogdon
 *
 *  Terms of use:  This software is provided for use under the terms
 *                 and conditions of the GNU General Public License.
 *                 You should have received a copy of the GNU General
 *                 Public License along with this program; if not, write
 *                 to the Free Software Foundation, Inc., 59 Temple Place
 *                 Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Global Data:  None.
 *  Global Functions:
 *    gdb_serial_init
 *    gdb_target_init
 *    gdb_target_close
 *    putDebugChar
 *    getDebugChar
 *
 *  History
 *  Engineer:           Date:              Notes:
 *  ---------           -----              ------
 *  Jonathan Brogdon    20000617           Genesis
 *  Gordon Schumacher   20020212           Updated for modularity 
 *
 ***********************************************************************/

#if defined(REMDEBUG)

#include <bios.h>
#include <stdlib.h>
#include "i386-stub.h"

//#include "bios_layer.h"			// Include this for BIOS calls - NOT RECOMMENDED!
//#include "sva_layer.h"			// Include this for SVAsync usage
//#include "dzc_layer.h"			// Include this for DZComm usage

#define SER_TIMEOUT	1000000

/***********************************************************************
 *  gdb_serial_init
 *
 *  Description:  Initializes the serial port for remote debugging.
 *
 *  Inputs:
 *    port        - the PC COM port to use.
 *    speed       - the COM port speed.
 *  Outputs:  None.
 *  Returns:  0 for success
 *
 ***********************************************************************/
int gdb_serial_init(unsigned int port, unsigned int speed)
{
	int ret;

	ret = GDBStub_SerInit(port);
	if (ret == 0)
		ret = GDBStub_SerSpeed(speed);

	return ret;
}

/***********************************************************************
 *  gdb_target_init
 *
 *  Description:  This function inializes the GDB target.
 *
 *  Inputs:   None.
 *  Outputs:  None.
 *  Returns:  None.
 *
 ***********************************************************************/
void gdb_target_init(void)
{
	set_debug_traps();
	atexit(restore_traps);
}

/***********************************************************************
 *  gdb_target_close
 *
 *  Description:  This function closes the GDB target.
 *
 *  Inputs:   None.
 *  Outputs:  None.
 *  Returns:  None.
 *
 ***********************************************************************/
void gdb_target_close(void)
{
	restore_traps();
}

/***********************************************************************
 *  putDebugChar
 *
 *  Description:  sends a character to the debug COM port.
 *
 *  Inputs:
 *    c           - the data character to be sent
 *  Outputs:  None.
 *  Returns:  0 for success
 *
 ***********************************************************************/
int putDebugChar(char c)
{
	register int timeout = 0;

	while ((GDBStub_SerSendOk() == 0) && (timeout < SER_TIMEOUT))
		timeout++;
	return GDBStub_SerSend(c);
}

/***********************************************************************
 *  getDebugChar
 *
 *  Description:  gets a character from the debug COM port.
 *
 *  Inputs:   None.
 *  Outputs:  None.
 *  Returns:  character data from the serial support.
 *
 ***********************************************************************/
int getDebugChar(void)
{
	register int ret = -1;
	register int timeout = 0;

	while ((GDBStub_SerRecvOk() == 0) && (timeout < SER_TIMEOUT))
		timeout++;
	if (timeout < SER_TIMEOUT)
		ret = GDBStub_SerRecv();

	return ret;
}

#endif // if defined(REMDEBUG)
